(require 'predicate-dispatch)
(use-package 'predicate-dispatch)

;; translations of examples from "Predicate Dispatching: A Unified
;; Theory of Dispatch" (Ernst, Kaplan, and Chambers) in ECOOP '98.

;;; p. 189
(defpdmethod zip (l1 l2)
  ((consp l1)
   (consp l2))
  (cons (cons (car l1) (car l2))
	(zip (cdr l1) (cdr l2))))

(defpdmethod zip (l1 l2)
  ((or (null l1) (null l2)))
  nil)

;;; p. 190; already handled by CLOS's multiple dispatching.
(defmethod zip2 ((l1 cons) (l2 cons))
  (cons (cons (car l1) (car l2))
	(zip2 (cdr l1) (cdr l2))))

(defmethod zip2 (l1 l2)
  nil)

;;; pp. 190--191
;;; they needn't worry so much about ambiguity -- anything nontrivial
;;; that applies is useful.
(defclass expr () ())

(defclass atomic-expr (expr) ())
(defclass var-ref (atomic-expr) ()) ;; would have slots in practice.

(defclass int-const (atomic-expr)
  ((value :reader value-of)))

(defclass binop () ())
(defclass int-plus (binop) ())
(defclass int-mul (binop) ())

(defclass binop-expr (expr)
  ((op :reader op-of)
   (arg1 :reader arg1-of)
   (arg2 :reader arg2-of)))

(defpdmethod constant-fold (e)
  () ;; default method.
  e)

(defpdmethod constant-fold ((e binop-expr))
  ((typep (op-of e) 'int-plus)
   (typep (arg1-of e) 'int-const)
   (typep (arg2-of e) 'int-const))
  (make-instance 'int-const :value (+ (value-of (arg1-of e))
				      (value-of (arg2-of e)))))

(defpdmethod constant-fold ((e binop-expr))
  ((typep (op-of e) 'int-mul)
   (typep (arg1-of e) 'int-const)
   (typep (arg2-of e) 'int-const))
  (make-instance 'int-const :value (* (value-of (arg1-of e))
				      (value-of (arg2-of e)))))

#||
;; &aux isn't quite as flexible as their syntax :-/
(defpdmethod constant-fold ((e binop-expr)
			    &aux
			    (v (value-of (arg1-of e)))
			    (a2 (arg2-of e)))
  ((typep (op-of e) 'int-plus)
   (typep (arg1-of e) 'int-const)
   (zerop v)
   (not (typep a2 'int-const))) ;; avoid possible ambiguity
  a2)

(defpdmethod constant-fold ((e binop-expr)
			    &aux
			    (a1 (arg1-of e))
			    (v (value-of (arg2-of e))))
  ((typep (op-of e) 'int-plus)
   (not (typep a1 'int-const)) ;; avoid possible ambiguity
   (typep (arg2-of e) 'int-const)
   (zerop v))
  a1)
||#

(defpdmethod constant-fold ((e binop-expr) &aux (a2 (arg2-of e)))
  ((typep (op-of e) 'int-plus)
   (typep (arg1-of e) 'int-const)
   (zerop (value-of (arg1-of e))) ;; guarded by previous check
   (not (typep a2 'int-const))) ;; avoid possible ambiguity
  a2)

(defpdmethod constant-fold ((e binop-expr) &aux (a1 (arg1-of e)))
  ((typep (op-of e) 'int-plus)
   (not (typep a1 'int-const)) ;; avoid possible ambiguity
   (typep (arg2-of e) 'int-const)
   (zerop (value-of (arg2-of e))) ;; guarded by previous check
  a1)

;; etc. etc. etc.

;;; p. 192, various approaches
(defmacro on-x-axis-1 (p) ;; unlike a function, permits subanalysis
  `(or (and (typep ,p cartesian-point) (zerop (y ,p)))
       (and (typep ,p polar-point)
	    (or (zerop (theta ,p))
		(= (theta ,p) pi)))))

(defpdmethod draw1 (p)
  ((on-x-axis-1 p))
  ;...
  )

(setf on-x-axis-2
      (build-predicate '(p)
		       '(or (and (typep p cartesian-point) (zerop (y p)))
			    ...)))
;; no syntax for using (yet)

;;; next would be classifiers -- macro wrapper for whichever approach?
