(require 'pd-package)
(in-package :predicate-dispatch)

;; Abstract transformations:
;; * p v ~p -> *true*
;; * (p+ ^ q*) v p+ -> p+
;; * (p ^ q+) v (~p ^ q+) -> q+
;; * (p ^ q* ^ r+) v (~p ^ q*) -> (q* ^ r+) v (~p ^ q*)
;; * (p ^ q* ^ r+) v (~p ^ q* ^ s+)
;;   -> (q* ^ r+ ^ s+) v (p ^ q* ^ r+) v (~p ^ q* ^ s+)
;;   [may be useful in setting the stage for other transformations]

;; prove complete?

;; interesting test cases:
;; * (p ^ q) v (p ^ ~q) v (~p ^ q) v (~p ^ ~q)
;; * (p ^ q ^ r) v ~p v ~q v ~r
;; * (p ^ q) v (~q ^ r) v (~r ^ ~p) v (p ^ ~q ^ ~r) v (~p ^ q ^ r)

(defun flattened-or-subpredicates (p)
  (if (typep p 'or-predicate)
      (apply #'append (mapcar #'flattened-or-subpredicates (subpreds-of p)))
    (list p)))

(defun extra-flattened-and-subpredicates (p)
  (cond ((typep p 'and-predicate)
	 (apply #'append (mapcar #'extra-flattened-and-subpredicates
				 (subpreds-of p))))
	((and (typep p 'projected-unary-predicate)
	      (typep (base-of p) 'and-predicate))
	 (mapcar #'(lambda (x) (make-instance 'projected-unary-predicate
					      :base x :index (index-of p)))
		 (subpreds-of (base-of p))))
	(t (list p))))

(defun maybe-make-eup (base accessors)
  (if accessors (make-instance 'extracting-unary-predicate
			       :base base :accessors accessors)
    base))

(defun maybe-flip (orig pairs flip?)
  (if flip?
      (or (cdr (assoc orig pairs))
	  (car (rassoc orig pairs)))
    orig))

(defun adjust-comparison (orig neg-x neg-y)
  (maybe-flip (maybe-flip orig
			  '((same . opposite)
			    (forward . comprehensive)
			    (backward . exclusive))
			  neg-x)
	      '((same . opposite)
		(forward . exclusive)
		(backward . comprehensive))
	      neg-y))

(defun compare-terms (x y &optional neg-x neg-y)
  (adjust-comparison
   (cond ((typep x 'projected-unary-predicate)
	  (and (typep y 'projected-unary-predicate)
	       (= (index-of x) (index-of y))
	       (compare-terms (base-of x) (base-of y) neg-x neg-y)))
	 ((typep y 'projected-unary-predicate) nil)
	 ;;
	 ((and (typep x 'equality-predicate)
	       (typep y 'equality-predicate))
	  (if (eql (target-of x) (target-of y)) 'same 'exclusive))
	 ;; no appropriate auxv...pass nil instead, and ignore errors.
	 ((typep x 'equality-predicate)
	  (ignore-errors
	    (if (evaluate-predicate y (target-of x) nil)
		'forward
	      'exclusive)))
	 ((typep y 'equality-predicate)
	  (ignore-errors
	    (if (evaluate-predicate x (target-of y) nil)
		'backward
	      'exclusive)))
	 ;;
	 ((typep x 'not-predicate)
	  (compare-terms (base-of x) y (not neg-x) neg-y))
	 ((typep y 'not-predicate)
	  (compare-terms x (base-of y) neg-x (not neg-y)))
	 ;;
	 ((and (typep x 'extracting-unary-predicate)
	       (typep y 'extracting-unary-predicate))
	  (do ((x-acc (accessors-of x) (cdr x-acc))
	       (y-acc (accessors-of y) (cdr y-acc)))
	      ((or (null x-acc) (null y-acc)
		   (not (eql (car x-acc) (car y-acc))))
	       (and (or (null x-acc) (null y-acc))
		    (compare-terms (maybe-make-eup (base-of x) x-acc)
				   (maybe-make-eup (base-of y) y-acc)
				   neg-x neg-y)))))
	 ;;
	 ((and (typep x 'typecheck-predicate)
	       (typep y 'typecheck-predicate))
	  (let ((tx (target-of x))
		(ty (target-of y)))
	    ;; In a more static environment, we'd want to compare the sets
	    ;; of concrete subtypes.  As it is, we have to be conservative.
	    (cond ((eql tx ty) 'same)
		  ((subtypep tx ty) 'forward)
		  ((subtypep ty tx) 'backward)
		  (t nil))))
	 ;;
	 ((and (typep x 'test-predicate)
	       (typep y 'test-predicate))
	  (if (eql (test-of x) (test-of y))
	      'same
	    nil))
	 ;;
	 (t nil))
   neg-x neg-y))

(defun analyze-term (x yy)
  (let ((comparisons (remove nil (mapcar #'(lambda (y) (compare-terms x y))
					 yy))))
    (cond ((null comparisons) 'extra)
	  ((member 'exclusive comparisons) 'fatal-mismatch)
	  ((member 'opposite comparisons) 'mismatch)
	  ((member 'same comparisons) 'match)
	  ;; ((member 'forward comparisons) 'match)
	  ((member 'backward comparisons) 'weak)
	  ;; "comprehensive" and "forward" don't help us.
	  (t 'extra))))

(defun annotate-term (x yy)
  (cons x (analyze-term x yy)))

(defun keep? (x y)
  (or (member 'extra y :key #'cdr)
      (member 'weak x :key #'cdr)))

(defun compute-safe-patch (left right)
  ;; Return values:
  ;; (1) Safe patch if appropriate (single mismatch), nil otherwise.
  ;; (2) Does left cover any extra territory?
  ;; (3) Does right?
  (let* ((fleft (extra-flattened-and-subpredicates left))
	 (fright (extra-flattened-and-subpredicates right))
	 (xleft (mapcar #'(lambda (l) (annotate-term l fright)) fleft))
	 (mismatches (count 'mismatch xleft :key #'cdr)))
    (if (or (member 'fatal-mismatch xleft :key #'cdr)
	    (> mismatches 1))
	(values nil t t) ;; We lose.
      (let* ((xright (mapcar #'(lambda (r) (annotate-term r fleft)) fright))
	     (keep-right (keep? xright xleft)))
	(if (zerop mismatches)
	    (values nil
		    (or (not keep-right)
			(keep? xleft xright))
		    keep-right)
	  (values
	   ;; Merge useful terms into a patch.  Keep weak terms on the left
	   ;; despite their redundancy because something may depend on them
	   ;; wrt short-circuiting.
	   ;; Return t instead of an empty patch.
	   (or (mapcar #'car
		       (delete 'mismatch
			       (nconc xleft
				      (delete-if
				       #'(lambda (x)
					   (member (cdr x) '(match weak)))
				       xright))
			       :key #'cdr))
	       t)
	   (keep? xleft xright)
	   keep-right))))))

(defmethod normalize-predicate ((predicate or-predicate))
  (cond ((normal? predicate) predicate)
	((null (subpreds-of predicate)) *false*)
	(t
	 (let ((subpreds (mapcar #'normalize-predicate
				 (flattened-or-subpredicates predicate))))
	   (if (some #'value-of (remove-if-not #'constant-predicate? subpreds))
	       *true*
	     ;; can mutate here because nothing else refers to subpreds.
	     (do ((queue (delete-if #'constant-predicate? subpreds)
			 (cdr queue))
		  (new-subpreds nil (if insert-head
					(cons (car queue) survivors)
				      survivors))
		  (survivors nil nil)
		  (insert-head t (and (not insert-head) (null survivors))))
		 ((null queue) (if (cdr new-subpreds)
				   (make-or new-subpreds t)
				 (car new-subpreds)))
	       ;; (format t "Queue:~:W~%NewS: ~:W~%" queue new-subpreds)
	       (dolist (subpred new-subpreds)
		 (multiple-value-bind (patch keep-left keep-right)
		     (compute-safe-patch (car queue) subpred)
		   ;; (format t "Patch:~:W~%KeepL:~:W~%KeepR:~:W~%" patch
		   ;;  keep-left keep-right)
		   (if patch
		       (if (eq patch t)
			   (return-from normalize-predicate *true*);; we win!
			 (push (normalize-predicate (make-and patch))
			       (cdr queue))))
		   (if keep-left
		       (setf insert-head t))
		   (if keep-right
		       (push subpred survivors))))))))))

(provide 'normalize-or)
