(in-package :weyli)
(require 'predicate-dispatch)
(use-package 'predicate-dispatch)

(defclass ge-integral (general-expression)
  ((expr :initarg :expr
	 :accessor expression-of)
   (var :initarg :var
	:accessor variable-of)
   (lower :initarg :lower
	  :initform nil
	  :accessor lower-bound-of)
   (upper :initarg :upper
	  :initform nil
	  :accessor upper-bound-of)))

(defmethod print-object ((int ge-integral) stream)
  (format stream "int ~A d~A" (expression-of int) (variable-of int))
  (if (lower-bound-of int)
      (format stream " from ~A to ~A" (lower-bound-of int)
	      (upper-bound-of int))))

(defpdmethod integral (expr var) ()
  (let ((d (typecase expr
	     (domain-element (domain-of expr))
	     (t *general*))))
    (make-instance 'ge-integral :expr expr :var var :domain d)))

(defun definite-integral (expr var lower upper)
  (let* ((ge-var (coerce var *general*))
	 (indef (integral expr ge-var)))
    (if (typep indef 'ge-integral)
	(make-instance 'ge-integral :expr expr :var var
		       :domain (domain-of indef)
		       :lower lower :upper upper)
      (- (substitute (coerce upper *general*) ge-var indef)
	 (substitute (coerce lower *general*) ge-var indef)))))

(defun constant-of-integration ()
  (coerce (gensym "C") *general*))

(defun indefinite-integral (expr var)
  (+ (integral expr (coerce var *general*))
     (constant-of-integration)))

(defun first-domain (&rest args)
  (cond ((null args) nil)
	((typep (car args) 'domain-element)
	 (domain-of (car args)))
	(t (apply #'first-domain (cdr args)))))

(defun int (expr var &optional lower upper)
  (if lower
      (weyli::definite-integral expr var lower upper)
    (weyli::indefinite-integral expr var)))

(defmethod symbol-of ((sym symbol))
  sym)

(defun free? (expr var)
  ;;(zerop (deriv expr var))
  (not (depends-on? expr var))
  )

(defun make-free-in? (var)
  (lambda (expr) (free? expr var)))

;; useful for a limited version of the chain rule
(defun linear? (expr var)
  (let ((d (deriv expr var)))
    (and (not (zerop d))
	 (free? d var))))

(defun same-var? (var1 var2)
  (and (or (symbolp var1) (typep var1 'ge-variable))
	(or (symbolp var2) (typep var2 'ge-variable))
	(eq (symbol-of var1) (symbol-of var2))))

(export 'int)

;;; useful specializations follow

(defpdmethod integral (expr var)
  ((free? expr var))
  (* expr var))

(defpdmethod integral (expr var)
  ((same-var? expr var))
  (* 1/2 expr expr))

(defpdmethod integral ((expr ge-plus) var) ()
  (make-ge-plus (domain-of expr)
		(mapcar (lambda (exp)
			  (integral exp var))
			(terms-of expr))))

(defpdmethod integral ((expr ge-times) var
		       &aux (free? (make-free-in? var))
		       (terms (terms-of expr)))
  ((member-if free? terms))
  (let ((domain (domain-of expr)))
    (make-ge-times domain
		   (cons (integral (make-ge-times domain
						  (remove-if free? terms))
				   var)
			 (remove-if-not free? terms)))))

(defpdmethod integral ((expr ge-times) var)
  ((= (length (terms-of expr)) 1))
  (integral (car (terms-of expr)) var))

(defpdmethod integral ((expr ge-expt) var
		       &aux (base (base-of expr)) (exp (exponent-of expr)))
  ((free? exp var)
   (linear? base var))
  (let ((exp1 (+ exp 1)))
    (if (zerop exp1)
	;; XXX -- should take absolute value of log.
	(/ (make-ge-log (domain-of expr) base) (deriv base var))
      (/ (make-ge-expt (domain-of expr) base exp1) exp1 (deriv base var)))))

(defpdmethod integral ((expr ge-expt) var
		       &aux (base (base-of expr)) (exp (exponent-of expr)))
  ((free? base var)
   (linear? exp var))
  (if (= base 1)
      (/ exp (deriv base var))
    (/ (make-ge-expt (domain-of expr) base exp) (log base) (deriv base var))))
   
(defpdmethod integral ((expr ge-application) var
		       &aux (first-arg (first (args-of expr))))
  ((equal (name-of (funct-of expr)) "sin")
   (linear? first-arg var))
  (- (/ (make-ge-cos (domain-of expr) first-arg) (deriv first-arg var))))

(defpdmethod integral ((expr ge-application) var
		       &aux (first-arg (first (args-of expr))))
  ((equal (name-of (funct-of expr)) "cos")
   (linear? first-arg var))
  (/ (make-ge-sin (domain-of expr) first-arg) (deriv first-arg var)))

(defpdmethod integral ((expr ge-application) var
		       &aux (first-arg (first (args-of expr))))
  ((equal (name-of (funct-of expr)) "tan")
   (linear? first-arg var))
  (- (/ (make-ge-log (domain-of expr)
		     (make-ge-cos (domain-of expr) first-arg))
	(deriv first-arg var))))

(defpdmethod integral ((expr ge-application) var
		       &aux (first-arg (first (args-of expr))))
  ((equal (name-of (funct-of expr)) "log")
   (linear? first-arg var))
  (/ (- (* first-arg (make-ge-log (domain-of expr) first-arg))
	first-arg)
     (deriv first-arg var)))

(defpdmethod integral ((expr ge-application) var
		       &aux (first-arg (first (args-of expr))))
  ((equal (name-of (funct-of expr)) "sinh")
   (linear? first-arg var))
  (/ (make-ge-cosh (domain-of expr) first-arg) (deriv first-arg var)))

(defpdmethod integral ((expr ge-application) var
		       &aux (first-arg (first (args-of expr))))
  ((equal (name-of (funct-of expr)) "cosh")
   (linear? first-arg var))
  (/ (make-ge-sinh (domain-of expr) first-arg) (deriv first-arg var)))

(defpdmethod integral ((expr ge-application) var
		       &aux (first-arg (first (args-of expr))))
  ((equal (name-of (funct-of expr)) "tanh")
   (linear? first-arg var))
  (/ (make-ge-log (domain-of expr)
		  (make-ge-cosh (domain-of expr) first-arg))
     (deriv first-arg var)))

;; ...
