(require 'pd-package)
(in-package :predicate-dispatch)

;; proposed schema for normalized predicates:
;;
;; full multipredicate = PCM | or(PCM{2,}) | constant
;; full unary predicate [predicate class] = PCU | or(PCU{2,}) | constant
;;
;; PCM = SM | and(SM{2,})
;;  SM = test | not(test) | projected-unary(PCU)
;; PCU = PNU | and(PNU{2,})
;; PNU = PEU | not(PEU)
;; PEU = SU | extracting-unary(SU)
;;  SU = typecheck | equality | test
;;
;; abbreviations stand for (Purely Conjunctive)/Simple Multi-arg/Unary and
;; Possibly Negated/Extracting Unary.

(defclass predicate () ; abstract
  ((normal? :initarg :normal?
	    :reader normal?
	    :initform nil)))

(defclass test-predicate (predicate)
  ((test :initarg :test
	 :reader test-of)
   (pass-auxv :initarg :pass-auxv
	      :reader pass-auxv?
	      :initform t)))

(defclass constant-predicate (predicate)
  ((value :initarg :value
	  :reader value-of)))

(defclass typecheck-predicate (predicate)
  ((target-type :initarg :target
		:reader target-of)))

(defclass equality-predicate (predicate)
  ((target-value :initarg :target
		 :reader target-of)))

(defclass modified-predicate (predicate) ; abstract
  ((base-predicate :initarg :base
		   :reader base-of)))

(defclass projected-unary-predicate (modified-predicate)
  ((argument-index :initarg :index
		   :reader index-of)))

(defclass extracting-unary-predicate (modified-predicate)
  ((accessor-chain :initarg :accessors
		   :reader accessors-of)))

(defclass not-predicate (modified-predicate)
  ())

(defclass compound-predicate (predicate) ; abstract
  ((subpreds :initarg :subpreds
	     :reader subpreds-of)))

(defclass and-predicate (compound-predicate)
  ())

(defclass or-predicate (compound-predicate)
  ())

(defclass predicate-qualifier ()
  ((predicate :initarg :predicate
	      :reader predicate-of)
   (auxv-count :initarg :auxv-count
	       :reader auxv-count-of)))

(defun predicate-qualifier? (x)
  (typep x 'predicate-qualifier))

(defconstant *true* (make-instance 'constant-predicate :value t))
(defconstant *false* (make-instance 'constant-predicate :value nil))

(defmethod evaluate-predicate ((predicate test-predicate) args auxv)
  (values (apply (test-of predicate)
		 (if (pass-auxv? predicate) (cons auxv args) args))
	  auxv))

(defmethod evaluate-predicate ((predicate constant-predicate) args auxv)
  (declare (ignore args))
  (values (value-of predicate) auxv))

(defmethod evaluate-predicate ((predicate typecheck-predicate) args auxv)
  (values (typep (car args) (target-of predicate)) auxv))

(defmethod evaluate-predicate ((predicate equality-predicate) args auxv)
  (values (eql (car args) (target-of predicate)) auxv))

(defmethod evaluate-predicate ((predicate projected-unary-predicate) args auxv)
  (evaluate-predicate (base-of predicate)
		      (list (nth (index-of predicate) args)) auxv))

(defmethod evaluate-predicate ((predicate extracting-unary-predicate) args
			       auxv)
  (do ((arg (car args) (funcall (car accessors) arg))
       (accessors (accessors-of predicate) (cdr accessors)))
      ((null accessors) (evaluate-predicate (base-of predicate)
					    (list arg) auxv))))

(defmethod evaluate-predicate ((predicate and-predicate) args auxv)
  (values (every #'(lambda (subpred) (evaluate-predicate subpred args auxv))
		 (subpreds-of predicate))
	  auxv))

(defmethod evaluate-predicate ((predicate or-predicate) args auxv)
  (values (some #'(lambda (subpred) (evaluate-predicate subpred args
							;(copy-seq auxv)
							auxv))
		(subpreds-of predicate))
	  auxv))

(defmethod evaluate-predicate ((predicate not-predicate) args auxv)
  (values (not (evaluate-predicate (base-of predicate) args auxv)) auxv))

(defun make-or (subpreds &optional normal?)
  (make-instance 'or-predicate :subpreds subpreds :normal? normal?))

(defun make-and (subpreds &optional normal?)
  (make-instance 'and-predicate :subpreds subpreds :normal? normal?))

(defun make-not (base &optional normal?)
  (make-instance 'not-predicate :base base :normal? normal?))

(defun constant-predicate? (predicate)
  (typep predicate 'constant-predicate))

(defun or-predicate? (predicate)
  (typep predicate 'or-predicate))

(defmethod normalize-predicate (predicate)
  predicate)

(defmethod normalize-predicate ((predicate not-predicate))
  (if (normal? predicate)
      predicate
    (normalize-predicate
     (let ((base (normalize-predicate (base-of predicate))))
       (typecase base
	 (not-predicate (base-of base))
	 (and-predicate (make-or (mapcar #'make-not (subpreds-of base))))
	 (or-predicate (make-and (mapcar #'make-not (subpreds-of base))))
	 (projected-unary-predicate (make-instance 'projected-unary-predicate
						   :index (index-of base)
						   :base (make-not (base-of
								    base))))
	 (constant-predicate (make-instance 'constant-predicate
					    :value (not (value-of base))))
	 (otherwise (make-not base t)))))))

(require 'normalize-or) ;; split off due to size

(defun flattened-and-subpredicates (p)
  (if (typep p 'and-predicate)
      (apply #'append (mapcar #'flattened-and-subpredicates (subpreds-of p)))
    (list p)))

(defun safe-index-of (p)
  (and (typep p 'projected-unary-predicate)
       (index-of p)))

(defmethod normalize-predicate ((predicate and-predicate))
  (cond ((normal? predicate) predicate)
	((null (subpreds-of predicate)) *true*)
	((null (cdr (subpreds-of predicate)))
	 (normalize-predicate (car (subpreds-of predicate))))
	(t (let ((subpreds (mapcar #'normalize-predicate
				   (flattened-and-subpredicates predicate))))
	     (let ((first-or (find-if #'or-predicate? subpreds)))
	       (if first-or ;; distribute!
		   (normalize-predicate
		    (make-or
		     (mapcar #'(lambda (x)
				 (make-and
				  (substitute x first-or subpreds :count 1)))
			     (subpreds-of first-or))))
		 (do ((ht (make-hash-table))
		      (rsp (reverse subpreds) (cdr rsp)))
		     ((null rsp)
		      (let ((terms (gethash nil ht)))
			(maphash #'(lambda (index preds)
				     (if index
					 (push (make-instance
						'projected-unary-predicate
						:base (normalize-predicate
						       (make-and preds))
						:index index)
					       terms)))
				 ht)
			(make-and terms t)))
		   (if (typep (car rsp) 'projected-unary-predicate)
		       (push (base-of (car rsp))
			     (gethash (index-of (car rsp)) ht))
		     (push (car rsp) (gethash nil ht))))))))))

(defmethod normalize-predicate ((predicate projected-unary-predicate))
  (if (normal? predicate)
      predicate
    (flet ((make-cousin (new-base &optional normal?)
			(make-instance 'projected-unary-predicate
				       :base new-base
				       :index (index-of predicate)
				       :normal? normal?)))
      (let ((base (normalize-predicate (base-of predicate))))
	(typecase base
	  (or-predicate (normalize-predicate
			 (make-or (mapcar #'make-cousin (subpreds-of base)))))
	  (constant-predicate base)
	  (otherwise (make-cousin base t)))))))

(defmethod normalize-predicate ((predicate extracting-unary-predicate))
  (if (normal? predicate)
      predicate
    (normalize-predicate
     (flet ((make-cousin (new-base &optional normal?)
			 (make-instance 'extracting-unary-predicate
					:base new-base
					:accessors (accessors-of predicate)
					:normal? normal?)))
       (let ((base (normalize-predicate (base-of predicate))))
	 (typecase base
	   (or-predicate (make-or (mapcar #'make-cousin (subpreds-of base))))
	   (and-predicate (make-and (mapcar #'make-cousin (subpreds-of base))))
	   (constant-predicate base)
	   (not-predicate (make-not (make-cousin (base-of base))))
	   (otherwise (make-cousin base t))))))))

(defmethod implies? (pred1 pred2)
  ;; takes advantage of simplification done in normalize-or.lisp
  (let ((norm (normalize-predicate (make-or (make-not pred1) pred2))))
    (and (constant-predicate? norm) (value-of norm))))

(provide 'predicate-classes)
